"""Common compound type aliases."""

from collections import Counter
from dataclasses import dataclass
from importlib import import_module
from typing import TYPE_CHECKING, Any, Dict, Tuple

from .circuit import CircuitWrapper, UnitID

if TYPE_CHECKING:
    from pytket.backends.backendresult import BackendResult

    # use obscure names for following imports to make linting happy
    from pytket.partition import MeasurementBitMap as _PMBP
    from pytket.partition import MeasurementSetup as _PMS
    from pytket.pauli import QubitPauliString as PytketQubitPauliString

Distribution = Dict[str, float]


def _try_pytket_import(mod: str, obj: str) -> Any:
    try:
        module = import_module(mod)
        loaded = getattr(module, obj)
    except ModuleNotFoundError as e:
        raise ImportError("This function requires pytket installed.") from e

    return loaded


QubitPauliString = list[tuple[UnitID, str]]


def _qps_to_pytket(qps: QubitPauliString) -> "PytketQubitPauliString":
    PytketQubitPauliString = _try_pytket_import("pytket.pauli", "QubitPauliString")
    # Pauli = _try_pytket_import("pytket.pauli", "Pauli")
    # Qubit = _try_pytket_import("pytket.circuit", "Qubit")
    if not qps:
        return PytketQubitPauliString()
    return PytketQubitPauliString.from_list(
        [[uid.to_serializable(), pauli] for uid, pauli in qps]
    )


def _qps_from_pytket(tkqps: "PytketQubitPauliString") -> QubitPauliString:
    lst = tkqps.to_list()
    if not lst:
        return []
    return [(UnitID.from_serializable(qb), pauli) for qb, pauli in lst]


@dataclass(frozen=True)
class MeasurementBitMap:
    circ_index: int
    bits: list[int]
    invert: bool

    def to_pytket(self) -> "_PMBP":
        PytketMeasurementBitMap = _try_pytket_import(
            "pytket.partition", "MeasurementBitMap"
        )
        return PytketMeasurementBitMap(self.circ_index, self.bits, self.invert)

    @classmethod
    def from_pytket(cls, py_map: "_PMBP") -> "MeasurementBitMap":
        return cls(py_map.circ_index, py_map.bits, py_map.invert)


@dataclass(frozen=True)
class MeasurementSetup:
    measurement_circs: list[CircuitWrapper]  # using the wrapper to avoid pytket import
    results: list[tuple[QubitPauliString, list[MeasurementBitMap]]]

    def to_pytket(self) -> "_PMS":
        PytketMeasurementSetup = _try_pytket_import(
            "pytket.partition", "MeasurementSetup"
        )
        ms = PytketMeasurementSetup()
        for circ in self.measurement_circs:
            ms.add_measurement_circuit(circ.from_tierkreis_compatible())
        for qps, mbms in self.results:
            tkqps = _qps_to_pytket(qps)
            for mbm in mbms:
                ms.add_result_for_term(tkqps, mbm.to_pytket())

        return ms

    @classmethod
    def from_pytket(cls, py_map: "_PMS") -> "MeasurementSetup":
        return cls(
            [
                CircuitWrapper.new_tierkreis_compatible(c)
                for c in py_map.measurement_circs
            ],
            [
                (
                    _qps_from_pytket(tkqps),
                    [MeasurementBitMap.from_pytket(tkmbm) for tkmbm in tkmbms],
                )
                for tkqps, tkmbms in py_map.results.items()
            ],
        )



"""Utilities for building tierkreis graphs."""
import copy
import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

import betterproto
import networkx as nx

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
from tierkreis.core.function import FunctionName
from tierkreis.core.types import TierkreisType
from tierkreis.core.values import T, TierkreisValue

if TYPE_CHECKING:
    from tierkreis.builder import Unpack, ValueSource

PortID = str
Location = pg.Location


class TierkreisNode(ABC):
    @abstractmethod
    def to_proto(self) -> pg.Node:
        pass

    @classmethod
    def from_proto(cls, node: pg.Node) -> "TierkreisNode":
        name, out_node = betterproto.which_one_of(node, "node")

        result: TierkreisNode
        if name == "const":
            result = ConstNode(TierkreisValue.from_proto(cast(pg.Value, out_node)))
        elif name == "box":
            box_node = cast(pg.BoxNode, out_node)
            result = BoxNode(
                graph=TierkreisGraph.from_proto(box_node.graph), location=box_node.loc
            )
        elif name == "function":
            fn_node = cast(pg.FunctionNode, out_node)
            result = FunctionNode(
                FunctionName.from_proto(fn_node.name), fn_node.retry_secs
            )
        elif name == "input":
            result = InputNode()
        elif name == "output":
            result = OutputNode()
        elif name == "tag":
            result = TagNode(cast(str, out_node))
        elif name == "match":
            result = MatchNode()
        else:
            raise ValueError(f"Unknown protobuf node type: {name}")

        if not isinstance(result, cls):
            raise TypeError()

        return result

    def is_discard_node(self) -> bool:
        """Delete nodes have some special behaviour, check for it."""
        return False

    def is_copy_node(self) -> bool:
        return False

    def is_unpack_node(self) -> bool:
        return False


@dataclass(frozen=True)
class InputNode(TierkreisNode):
    def to_proto(self) -> pg.Node:
        return pg.Node(input=pg.Empty())


@dataclass(frozen=True)
class OutputNode(TierkreisNode):
    def to_proto(self) -> pg.Node:
        return pg.Node(output=pg.Empty())


@dataclass(frozen=True)
class ConstNode(TierkreisNode):
    value: TierkreisValue

    def to_proto(self) -> pg.Node:
        return pg.Node(const=self.value.to_proto())


@dataclass(frozen=True)
class BoxNode(TierkreisNode):
    graph: "TierkreisGraph"
    location: Location

    def to_proto(self) -> pg.Node:
        return pg.Node(box=pg.BoxNode(loc=self.location, graph=self.graph.to_proto()))


@dataclass(frozen=True)
class FunctionNode(TierkreisNode):
    function_name: FunctionName
    retry_secs: Optional[int] = None  # Stored as uint32, so always >=0

    def to_proto(self) -> pg.Node:
        return pg.Node(
            function=pg.FunctionNode(
                name=self.function_name.to_proto(), retry_secs=self.retry_secs
            )
        )

    def is_discard_node(self) -> bool:
        return self.function_name == FunctionName("discard")

    def is_copy_node(self) -> bool:
        return self.function_name == FunctionName("copy")

    def is_unpack_node(self) -> bool:
        return self.function_name == FunctionName("unpack_struct")


@dataclass(frozen=True)
class TagNode(TierkreisNode):
    tag_name: str

    def to_proto(self) -> pg.Node:
        return pg.Node(tag=self.tag_name)


@dataclass(frozen=True)
class MatchNode(TierkreisNode):
    def to_proto(self) -> pg.Node:
        return pg.Node(match=pg.Empty())


@dataclass(frozen=True)
class NodeRef(Iterable):
    idx: int
    graph: "TierkreisGraph"
    outports: Optional[list[str]] = field(default=None, compare=False)

    def with_retry_secs(self, retry_secs: int) -> "NodeRef":
        """If this NodeRef refers to a FunctionNode, set its retry time in seconds.
        Otherwise, raises ValueError."""
        node = self.graph[self.idx]
        if not isinstance(node, FunctionNode):
            raise ValueError(
                "{self} must refer to a FunctionNode, but actually refers to {node}"
            )
        self.graph[self.idx] = FunctionNode(node.function_name, retry_secs)
        return self

    def default_nodeport(self, create: bool = True) -> "NodePort":
        if self.outports:
            if len(self.outports) != 1:
                raise ValueError(f"Cannot choose one outport from {self.outports}")
            return self[self.outports[0]]
        return self["value"]

    def __getitem__(self, key: str) -> "NodePort":
        # syntactic sugar for creating references to output ports from node
        return NodePort(self, key)

    def __len__(self) -> int:
        return len(self.outports or [])

    def __iter__(self) -> Iterator["NodePort"]:
        return (self[p] for p in (self.outports or []))


@dataclass(frozen=True)
class NodePort:
    node_ref: NodeRef
    port: PortID

    def __getitem__(self, field_name: str) -> "Unpack":
        raise NotImplementedError(
            "Import tierkreis.builder to get"
            "automatic unpacking of structs using the [<field>] syntax. "
        )


@dataclass(frozen=True)
class TierkreisEdge:
    source: NodePort
    target: NodePort
    type_: Optional[TierkreisType]

    def to_proto(self) -> pg.Edge:
        edge_type = None if self.type_ is None else self.type_.to_proto()
        return pg.Edge(
            port_from=self.source.port,
            node_from=self.source.node_ref.idx,
            port_to=self.target.port,
            node_to=self.target.node_ref.idx,
            edge_type=cast(pg.Type, edge_type),
        )

    def to_edge_handle(self) -> "_EdgeData":
        return (
            self.source.node_ref.idx,
            self.target.node_ref.idx,
            (self.source.port, self.target.port),
        )


# allow specifying input as a port, node with single output
IncomingWireType = Union[NodePort, NodeRef]


def to_nodeport(n: IncomingWireType) -> NodePort:
    return n if isinstance(n, NodePort) else n.default_nodeport()


_EdgeData = Tuple[int, int, Tuple[str, str]]


def _to_edgedata(edgeit: Iterable[Any]) -> Iterator[_EdgeData]:
    return (
        (src, dest, (cast(str, srcport), cast(str, destport)))
        for (src, dest, (srcport, destport)) in edgeit
    )


class MismatchedGraphs(Exception):
    def __init__(self, graph: "TierkreisGraph", endpoint: NodeRef):
        self.graph = graph
        self.endpoint = endpoint
        super().__init__("Cannot add an edge whose endpoint is on a different graph")


class TierkreisGraph:
    """TierkreisGraph."""

    input_node_idx: int = 0
    output_node_idx: int = 1

    @dataclass(frozen=True)
    class MissingEdge(Exception):
        source: NodePort
        target: NodePort

    def __init__(self, name: str = "") -> None:
        self.name = name
        self._graph = nx.MultiDiGraph()
        inp = self.add_node(InputNode())
        assert inp.idx == self.input_node_idx
        output = self.add_node(OutputNode())
        assert output.idx == self.output_node_idx

        self.input_order: list[str] = []
        self.output_order: list[str] = []

    @property
    def input(self) -> NodeRef:
        return NodeRef(self.input_node_idx, self)

    @property
    def output(self) -> NodeRef:
        return NodeRef(self.output_node_idx, self)

    def inputs(self) -> List[str]:
        return [edge.source.port for edge in self.out_edges(self.input)]

    def outputs(self) -> List[str]:
        return [edge.target.port for edge in self.in_edges(self.output)]

    @property
    def n_nodes(self) -> int:
        return len(self._graph)

    def add_node(
        self,
        _tk_node: TierkreisNode,
        /,
        **incoming_wires: IncomingWireType,
    ) -> NodeRef:
        node_ref = NodeRef(self._graph.number_of_nodes(), self)
        self._graph.add_node(node_ref.idx, node_info=_tk_node)
        for target_port_name, source in incoming_wires.items():
            self.add_edge(source, node_ref[target_port_name])

        return node_ref

    def add_func(
        self,
        _tk_function: Union[str, FunctionName],
        _retry_secs: Optional[int] = None,
        /,
        **kwargs: IncomingWireType,
    ) -> NodeRef:
        f_name: FunctionName = (
            FunctionName.parse(_tk_function)
            if isinstance(_tk_function, str)
            else _tk_function
        )

        return self.add_node(FunctionNode(f_name, _retry_secs), **kwargs)

    def add_const(self, value: Any) -> NodeRef:
        return self.add_node(ConstNode(TierkreisValue.from_python(value)))

    def add_box(
        self,
        graph: "TierkreisGraph",
        /,
        **kwargs: IncomingWireType,
    ) -> NodeRef:
        # TODO restrict to graph i/o
        return self.add_node(BoxNode(graph, Location([])), **kwargs)

    def add_match(
        self,
        variant_value: IncomingWireType,
        /,
        **variant_handlers: IncomingWireType,
    ) -> NodeRef:
        return self.add_node(
            MatchNode(), variant_value=variant_value, **variant_handlers
        )

    def add_tag(
        self,
        tag: str,
        *,  # Following are keyword only
        value: IncomingWireType,
    ) -> NodeRef:
        return self.add_node(TagNode(tag), value=value)

    def insert_graph(
        self,
        graph: "TierkreisGraph",
        /,
        **kwargs: IncomingWireType,
    ) -> Dict[str, IncomingWireType]:
        """Given a graph and wires to connect to that graph's inputs,
        inline that graph into <self>, and return the outputs of
        the inserted graph."""

        index_map: Dict[int, int] = dict()
        for node_idx, node in enumerate(graph.nodes()):
            if node_idx not in {graph.input_node_idx, graph.output_node_idx}:
                index_map[node_idx] = self.add_node(node).idx

        return_outputs: Dict[str, IncomingWireType] = {}

        for edge in graph.edges():
            source_node = edge.source.node_ref.idx
            target_node = edge.target.node_ref.idx
            source_port = (
                kwargs[edge.source.port]
                if source_node == graph.input_node_idx
                else NodeRef(index_map[source_node], self)[edge.source.port]
            )

            if target_node == graph.output_node_idx:
                return_outputs[edge.target.port] = source_port
            else:
                target_port = NodeRef(index_map[target_node], self)[edge.target.port]

                self.add_edge(source_port, target_port, edge.type_)

        return return_outputs

    def inline_boxes(self, recursive=False) -> "TierkreisGraph":
        """Inline boxes by inserting the graphs they contain in to the parent
        graph. Optionally do this recursively.

        :return: Inlined graph
        :rtype: TierkreisGraph
        """

        graph = copy.deepcopy(self)
        deleted_nodes: list[int] = []
        # Iterate through self.nodes rather than graph.nodes so we can mutate latter
        for node_idx, node in enumerate(self.nodes()):
            if not isinstance(node, BoxNode):
                continue

            boxed_g = node.graph
            if recursive:
                boxed_g = boxed_g.inline_boxes(recursive)

            curr_inputs = graph.in_edges(node_idx)

            incoming_ports = {edge.target.port: edge.source for edge in curr_inputs}
            curr_outputs = [
                graph._to_tkedge(e)
                for e in _to_edgedata(graph._graph.out_edges(node_idx, keys=True))
            ]
            # Removal all incident edges
            graph._graph.remove_edges_from(
                list(graph._graph.out_edges(node_idx))
                + list(graph._graph.in_edges(node_idx))
            )

            deleted_nodes.append(node_idx)

            inserted_outputs = graph.insert_graph(boxed_g, **incoming_ports)

            for edge in curr_outputs:
                graph.add_edge(
                    inserted_outputs[edge.source.port], edge.target, edge.type_
                )
        if deleted_nodes:
            graph.remove_nodes(deleted_nodes)

        return graph

    def nodes(self) -> Iterator[TierkreisNode]:
        return (
            self._graph.nodes[idx]["node_info"]
            for idx in range(self._graph.number_of_nodes())
        )

    def edges(self) -> Iterator[TierkreisEdge]:
        return map(self._to_tkedge, _to_edgedata(self._graph.edges(keys=True)))

    def __getitem__(self, key: Union[int, NodeRef]) -> TierkreisNode:
        name = key.idx if isinstance(key, NodeRef) else key
        return self._graph.nodes[name]["node_info"]

    def __setitem__(self, key: Union[int, NodeRef], node: TierkreisNode):
        name = key.idx if isinstance(key, NodeRef) else key
        self._graph.nodes[name]["node_info"] = node

    def add_edge(
        self,
        source: IncomingWireType,
        node_port_to: NodePort,
        edge_type: Optional[Union[Type, TierkreisType]] = None,
    ) -> TierkreisEdge:
        tk_type = _to_tierkreis_type(edge_type)
        node_port_from = to_nodeport(source)
        if node_port_from.node_ref.graph is not self:
            raise MismatchedGraphs(self, node_port_from.node_ref)
        if node_port_to.node_ref.graph is not self:
            raise MismatchedGraphs(self, node_port_to.node_ref)

        # if port is currently connected to discard, replace that edge
        existing_edge = self.out_edge_from_port(node_port_from)
        if isinstance(existing_edge, TierkreisType):
            tk_type = existing_edge
        elif existing_edge is not None:
            raise ValueError(
                f"Already an edge from {node_port_from} to {existing_edge.target}"
            )
        edge_data = (
            node_port_from.node_ref.idx,
            node_port_to.node_ref.idx,
            (node_port_from.port, node_port_to.port),
        )
        self._graph.add_edge(
            *edge_data,
            type=tk_type,
        )

        return self._to_tkedge(edge_data)

    def remove_edge(self, edge: TierkreisEdge):
        assert edge.source.node_ref.graph is self
        assert edge.target.node_ref.graph is self
        self._graph.remove_edge(
            edge.source.node_ref.idx,
            edge.target.node_ref.idx,
            (edge.source.port, edge.target.port),
        )

    def remove_nodes(self, nodes: Iterable[int]):
        self._graph.remove_nodes_from(nodes)
        # shift node indices to account for missing nodes
        # without this step indices will not be contiguous
        nx.relabel_nodes(
            self._graph,
            {n: i for i, n in enumerate(sorted(self._graph.nodes()))},
            copy=False,
        )

    def annotate_input(
        self, input_port: str, edge_type: Optional[Union[Type, TierkreisType]]
    ):
        (in_edge,) = [
            e
            for e in _to_edgedata(self._graph.out_edges(self.input_node_idx, keys=True))
            if e[2][0] == input_port
        ]
        tk_type = _to_tierkreis_type(edge_type)
        self._graph.edges[in_edge]["type"] = tk_type

    def annotate_output(
        self, output_port: str, edge_type: Optional[Union[Type, TierkreisType]]
    ):
        (out_edge,) = [
            e
            for e in _to_edgedata(self._graph.in_edges(self.output_node_idx, keys=True))
            if e[2][1] == output_port
        ]

        tk_type = _to_tierkreis_type(edge_type)
        self._graph.edges[out_edge]["type"] = tk_type

    def get_edge(self, source: NodePort, target: NodePort) -> TierkreisEdge:
        try:
            return next(
                e for e in self.out_edges(source.node_ref) if e.target == target
            )
        except StopIteration as e:
            raise self.MissingEdge(source, target) from e

    def set_outputs(self, **kwargs: IncomingWireType) -> None:
        for out_name, port in kwargs.items():
            self.add_edge(port, self.output[out_name])

    def _to_tkedge(self, handle: _EdgeData) -> TierkreisEdge:
        src, tgt, (src_port, tgt_port) = handle
        edge_data = cast(
            Mapping, self._graph.get_edge_data(src, tgt, (src_port, tgt_port))
        )
        return TierkreisEdge(
            NodeRef(src, self)[src_port],
            NodeRef(tgt, self)[tgt_port],
            edge_data["type"],
        )

    def in_edges(self, node: Union[NodeRef, int]) -> Iterator[TierkreisEdge]:
        node_name = node if isinstance(node, int) else node.idx
        return map(
            self._to_tkedge, _to_edgedata(self._graph.in_edges(node_name, keys=True))
        )

    def out_edges(self, node: Union[NodeRef, int]) -> Iterator[TierkreisEdge]:
        node_idx = node if isinstance(node, int) else node.idx
        return map(
            self._to_tkedge,
            _to_edgedata(self._graph.out_edges(node_idx, keys=True)),
        )

    def discard(self, out_port: NodePort) -> None:
        _ = self.add_func("discard", value=out_port)

    def make_pair(
        self, first_port: IncomingWireType, second_port: IncomingWireType
    ) -> NodePort:
        make_n = self.add_func("make_pair", first=first_port, second=second_port)
        return make_n["pair"]

    def unpack_pair(self, pair_port: IncomingWireType) -> Tuple[NodePort, NodePort]:
        unp_n = self.add_func("unpack_pair", pair=pair_port)
        return unp_n["first"], unp_n["second"]

    def make_vec(self, element_ports: List[IncomingWireType]) -> NodePort:
        vec_port = self.add_const(list())["value"]

        for port in element_ports:
            vec_port = self.add_func("push", vec=vec_port, item=port)["vec"]

        return vec_port

    def vec_last_n_elems(self, vec_port: NodePort, n_elements: int) -> List[NodePort]:
        curr_arr = vec_port
        outports: List[NodePort] = []
        for _ in range(n_elements):
            pop = self.add_func("pop", vec=curr_arr)
            curr_arr = pop["vec"]
            outports.insert(0, pop["item"])
        return outports

    def copy_value(self, value: IncomingWireType) -> Tuple[NodePort, NodePort]:
        copy_n = self.add_func("copy", value=value)

        return copy_n["value_0"], copy_n["value_1"]

    def out_edge_from_port(self, source: NodePort) -> Optional[TierkreisEdge]:
        """
        If there is an edge at port, return it, else None.
        """
        out_edges = [
            out_edge
            for out_edge in self.out_edges(source.node_ref)
            if out_edge.source.port == source.port
        ]
        if len(out_edges) == 0:
            return None
        (out_edge,) = out_edges
        return out_edge

    def to_proto(self) -> pg.Graph:
        pg_graph = pg.Graph()
        pg_graph.nodes = [n.to_proto() for n in self.nodes()]
        pg_graph.edges = [e.to_proto() for e in self.edges()]
        pg_graph.name = self.name
        pg_graph.input_order = self.input_order
        pg_graph.output_order = self.output_order
        return pg_graph

    @classmethod
    def from_proto(cls, pg_graph: pg.Graph) -> "TierkreisGraph":
        tk_graph = cls()
        tk_graph.name = pg_graph.name
        tk_graph.input_order = pg_graph.input_order
        tk_graph.output_order = pg_graph.output_order

        for pgn in pg_graph.nodes[2:]:
            tk_graph.add_node(TierkreisNode.from_proto(pgn))

        for pg_edge in pg_graph.edges:
            source_node = NodeRef(pg_edge.node_from, tk_graph)
            target_node = NodeRef(pg_edge.node_to, tk_graph)
            source = NodePort(source_node, PortID(pg_edge.port_from))
            target = NodePort(target_node, PortID(pg_edge.port_to))
            # TODO make all edge type conversions possible
            edge_type = pg_edge.edge_type
            tk_edge_type: Optional[TierkreisType] = None
            if edge_type is not None:
                try:
                    tk_edge_type = TierkreisType.from_proto(edge_type)
                except ValueError:
                    tk_edge_type = None

            tk_graph.add_edge(source, target, tk_edge_type)
        return tk_graph

    def to_python(self, type_: typing.Type[T]) -> T:
        if isinstance(type_, typing.TypeVar) or type_ is TierkreisGraph:
            return cast(T, self)
        raise TypeError()

    def __call__(
        self: "TierkreisGraph", *args: "ValueSource", **kwargs: "ValueSource"
    ) -> NodeRef:
        raise NotImplementedError(
            "TierkreisGraph can only be called inside builder contexts."
        )


def _to_tierkreis_type(
    edge_type: Optional[Union[Type, TierkreisType]],
) -> Optional[TierkreisType]:
    if edge_type is None:
        return None
    else:
        return (
            edge_type
            if isinstance(edge_type, TierkreisType)
            else TierkreisType.from_python(edge_type)
        )


# GraphValue defined after TierkreisGraph to avoid circular/delayed import


@dataclass(frozen=True)
class GraphValue(TierkreisValue):
    value: TierkreisGraph
    _proto_name: str = "graph"

    @property
    def _instance_pytype(self) -> typing.Type:
        return TierkreisGraph

    def to_proto(self) -> pg.Value:
        return pg.Value(graph=self.value.to_proto())

    def _to_python_impl(self, type_: typing.Type[T]) -> T | None:
        from tierkreis.core.python import RuntimeGraph

        if isinstance(type_, typing.TypeVar):
            return cast(T, self)
        if typing.get_origin(type_) is RuntimeGraph:
            return cast(T, RuntimeGraph(self.value))
        if issubclass(type_, TierkreisGraph):
            return cast(T, self.value)

    @classmethod
    def from_proto(cls, value: Any) -> "TierkreisValue":
        return cls(TierkreisGraph.from_proto(cast(pg.Graph, value)))

    def __str__(self) -> str:
        return "GraphValue"

    def viz_str(self) -> str:
        return "Graph"


# allow graph displays in jupyter notebooks
from tierkreis.core.graphviz import tierkreis_to_graphviz  # noqa: E402

setattr(
    TierkreisGraph,
    "_repr_svg_",
    lambda self: tierkreis_to_graphviz(self)._repr_svg_(),  # type: ignore
)



"""TierkreisFunction encapsulation of available function signatures"""
from dataclasses import dataclass, field
from typing import List

import tierkreis.core.protos.tierkreis.v1alpha1.graph as pg
import tierkreis.core.protos.tierkreis.v1alpha1.signature as ps


@dataclass
class FunctionName:
    name: str
    namespaces: List[str] = field(default_factory=list)

    @classmethod
    def from_proto(cls, pg_entry: pg.FunctionName) -> "FunctionName":
        return cls(
            pg_entry.name,
            pg_entry.namespaces,
        )

    def to_proto(self) -> pg.FunctionName:
        return pg.FunctionName(
            name=self.name,
            namespaces=self.namespaces,
        )

    @classmethod
    def parse(cls, to_parse: str) -> "FunctionName":
        atoms = to_parse.split("::")
        name = atoms.pop(-1)
        return cls(name, atoms)

    def __str__(self):
        return "::".join(self.namespaces + [self.name])


FunctionDeclaration = ps.FunctionDeclaration



from abc import ABC, abstractmethod
from typing import Any, Dict

from tierkreis.core.signature import Signature
from tierkreis.core.tierkreis_graph import TierkreisGraph
from tierkreis.core.values import TierkreisValue


class RuntimeClient(ABC):
    @abstractmethod
    async def get_signature(self) -> Signature:
        ...

    @abstractmethod
    async def run_graph(
        self,
        graph: TierkreisGraph,
        /,
        **py_inputs: Any,
    ) -> Dict[str, TierkreisValue]:
        ...

    @abstractmethod
    async def type_check_graph(self, graph: TierkreisGraph) -> TierkreisGraph:
        ...

    @property
    def can_type_check(self) -> bool:
        return True



import re

import pydantic as pyd

# regex pattern to find capital letters in a string (excluding the first character)
SNAKE_PATTERN: re.Pattern = re.compile(r"(?<!^)(?=[A-Z])")


def _to_snake_case(s: str) -> str:
    # insert underscore before capital letters and convert to lower case
    return SNAKE_PATTERN.sub("_", s).lower()


class OpaqueModel(pyd.BaseModel):
    """A type that is opaque to the Tierkreis system. Mapped to a StructType
    with a single field (field name may be customise by overloading
    `tierkreis_field`) with a string value holding the JSON serialized
    model."""

    @classmethod
    def tierkreis_field(cls) -> str:
        """Return the name of the field that will hold the JSON serialized string"""
        return "__tk_opaque_" + _to_snake_case(cls.__name__)



import inspect
import typing
from dataclasses import dataclass, fields, is_dataclass, replace
from typing import Any, Dict, ParamSpec, cast

from pydantic import BaseModel

from tierkreis.core.opaque_model import OpaqueModel


def substitute(
    type_: typing.Type, subst: Dict[typing.TypeVar, typing.Type]
) -> typing.Type:
    "Substitute type variables in a type."
    if isinstance(type_, typing.TypeVar):
        if type_ in subst:
            return subst[type_]
        else:
            return cast(typing.Type, type_)

    type_origin = typing.get_origin(type_)
    if type_origin is None:
        return type_
    else:
        return type_origin[  # type: ignore
            tuple([substitute(arg_type, subst) for arg_type in typing.get_args(type_)])
        ]


Out = tuple[typing.Type, str | None]


def _assert_annotation(annotation: typing.Type | None) -> typing.Type:
    if annotation is None:
        raise ValueError("Classes without type annotations cannot be converted.")

    return annotation


@dataclass(frozen=True)
class ClassField:
    """Capture required data about a class field."""

    name: str
    type_: typing.Type
    # the discriminant tag for the field, if it is a discriminated union
    discriminant: str | None
    # whether the field is present in the __init__ function
    init: bool = True
    default: Any = None


def python_struct_fields(
    type_: typing.Type | ParamSpec,
) -> list[ClassField]:
    """For a python dataclass or pydantic BaseModel"""
    type_origin = typing.get_origin(type_)
    if type_origin is not None:
        # Generic type
        type_subst = dict(zip(cast(Any, type_).__parameters__, typing.get_args(type_)))

        return [
            replace(field, type_=substitute(field.type_, type_subst))
            for field in python_struct_fields(type_origin)
        ]

    if inspect.isclass(type_) and issubclass(type_, BaseModel):
        import pydantic as pyd

        if issubclass(type_, OpaqueModel):
            model_type = cast(typing.Type[OpaqueModel], type_)

            return [
                ClassField(
                    model_type.tierkreis_field(),
                    str,
                    None,
                )
            ]
        model_type = cast(typing.Type[pyd.BaseModel], type_)
        model_fields = model_type.model_fields

        def _get_discriminator(f: pyd.fields.FieldInfo) -> str | None:
            disc = f.discriminator
            if isinstance(disc, pyd.Discriminator):
                raise ValueError("Discriminators must be static strings.")
            return disc

        return [
            ClassField(
                k,
                _assert_annotation(f.annotation),
                _get_discriminator(f),
                default=f.default,
            )
            for k, f in model_fields.items()
        ]
    if is_dataclass(type_):
        dat_fields = fields(type_)
        types = typing.get_type_hints(type_)
        return [
            ClassField(
                f.name,
                types[f.name],
                getattr(f.default, "discriminator", None),
                default=f.default,
            )
            for f in dat_fields
        ]

    raise ValueError("Can only convert dataclasses or pydantic BaseModel")



from dataclasses import dataclass
from typing import Any, Iterable, Optional

import pytest

from tierkreis.core import TierkreisGraph
from tierkreis.core.function import FunctionName
from tierkreis.core.tierkreis_graph import (
    BoxNode,
    FunctionNode,
    NodePort,
    NodeRef,
    TierkreisEdge,
)
from tierkreis.core.types import (
    IntType,
    StructType,
    TierkreisType,
    UnionTag,
    VariantType,
    VarType,
)
from tierkreis.core.values import (
    FloatValue,
    IntValue,
    MapValue,
    TierkreisValue,
    TierkreisVariant,
    VecValue,
)


def count(i: Iterable[Any]) -> int:
    return sum(1 for _ in i)


def test_creation_roundtrip() -> None:
    tg = TierkreisGraph()

    add = tg.add_func("iadd", a=tg.add_const(3), b=tg.input["input"])
    tg.set_outputs(output=add)

    id_g = TierkreisGraph()
    id_node = id_g.add_func("id", 2, value=id_g.input["value"])
    id_g.set_outputs(value=id_node["value"])

    b = tg.add_box(id_g)

    deser_tg = TierkreisGraph.from_proto(tg.to_proto())

    for graph in (tg, deser_tg):
        assert count(graph.nodes()) == 5
        assert count(graph.edges()) == 3

        assert graph.inputs() == ["input"]
        assert graph.outputs() == ["output"]

        add_node = NodeRef(add.idx, graph)
        assert graph[add_node] == FunctionNode(FunctionName.parse("iadd"))
        assert graph[add_node] != FunctionNode(FunctionName.parse("iadd"), 0)
        assert next(graph.out_edges(add)) == TierkreisEdge(
            add_node["value"], NodePort(graph.output, "output"), None
        )

        boxn = graph[b.idx]
        assert isinstance(boxn, BoxNode)
        assert boxn.graph[id_node.idx] == FunctionNode(FunctionName.parse("id"), 2)


def test_insert_subgraph() -> None:
    subgraph = TierkreisGraph()

    pair_port = subgraph.make_pair(subgraph.input["one"], subgraph.input["two"])
    first_p, second_p = subgraph.unpack_pair(pair_port)
    subgraph.discard(second_p)
    subgraph.set_outputs(sub_out=first_p)

    main_g = TierkreisGraph()
    two = main_g.add_const(3)
    main_nodes_before = main_g.n_nodes
    sub_nodes_before = subgraph.n_nodes
    subgraph_outs = main_g.insert_graph(subgraph, one=main_g.input["in1"], two=two)
    assert main_g.n_nodes == (sub_nodes_before + main_nodes_before) - 2

    assert list(subgraph_outs.keys()) == ["sub_out"]
    make_p = main_g.make_pair(main_g.input["in2"], subgraph_outs["sub_out"])

    main_g.set_outputs(value=make_p)

    assert count(main_g.nodes()) == 7
    assert count(main_g.edges()) == 7


def test_value_topython():
    convertible_vals = (
        1,
        "two",
        False,
        2.3,
        [1, 2],
        ("a", 4),
        {"asf": 3, "fsd": 4},
        TierkreisVariant("mylabel", 5),
        None,
    )

    for val in convertible_vals:
        assert TierkreisValue.from_python(val).try_autopython() == val

    fail_vals = ([1, 2.2], {"asd": 3, "fgh": False})
    for val in fail_vals:
        assert TierkreisValue.from_python(val).try_autopython() is None


@dataclass
class Foo:
    head: int
    tail: Optional["Foo"]


def test_type_from_python():
    t = TierkreisType.from_python(Foo)
    assert isinstance(t, StructType)
    fields = t.shape.content
    tail = fields.pop("tail")
    assert isinstance(tail, VariantType)
    some = tail.shape.content["__py_union_foo"]
    assert isinstance(some, VarType) and some.name.startswith("CyclicType")
    assert fields == {"head": IntType()}

    @dataclass
    class Bar:
        data1: Optional[int]
        data2: Optional[int]

    t = TierkreisType.from_python(Bar)
    assert isinstance(t, StructType)
    fields = t.shape.content
    d1 = fields.pop("data1")
    assert isinstance(d1, VariantType)
    assert d1.shape.content.keys() == frozenset(
        [UnionTag.type_tag(int), UnionTag.none_type_tag()]
    )
    assert d1.shape.content[UnionTag.type_tag(int)] == IntType()
    assert fields == {"data2": d1}


@pytest.mark.xfail(raises=AssertionError)
def test_value_topython_list_with_empty():
    pyval = [[], [3, 4]]
    tkval = TierkreisValue.from_python(pyval)
    if tkval != VecValue([VecValue([]), VecValue([IntValue(3), IntValue(4)])]):
        # We expect this to work (even though xfail'd) - else test setup has gone wrong
        pytest.fail()
    pyval2 = tkval.try_autopython()
    # The following fails because we can't figure out the element-type of the outer list
    # from the first element (the empty list). Looking at the second element would work.
    assert pyval2 == pyval


@pytest.mark.xfail(raises=AssertionError)
def test_value_topython_map_with_empty():
    pyval = {10: [[], [3, 4]], 15: [[2.1, 3.2], []]}
    tkval = TierkreisValue.from_python(pyval)
    if tkval != MapValue(
        {
            IntValue(10): VecValue.from_iterable(
                (VecValue([]), VecValue([IntValue(3), IntValue(4)]))
            ),
            IntValue(15): VecValue.from_iterable(
                (VecValue([FloatValue(2.1), FloatValue(3.2)]), VecValue([]))
            ),
        }
    ):
        # We expect this to work (even though xfail'd) - else test setup has gone wrong
        pytest.fail()
    pyval2 = tkval.try_autopython()
    # The following fails because we can't figure out the value-type of the Map
    # from either element (whichever comes first!).
    # We'd need to combine type info from both values.
    assert pyval == pyval2


def test_inline_boxes():
    tg_box = TierkreisGraph()

    pair = tg_box.make_pair(tg_box.input["inp"], tg_box.add_const(3))
    tg_box.set_outputs(out=pair)

    def check_inlined(graph: TierkreisGraph) -> bool:
        return any(
            isinstance(node, FunctionNode) and node.function_name.name == "make_pair"
            for node in graph.nodes()
        )

    tg = TierkreisGraph()
    box = tg.add_box(tg_box, inp=tg.add_const("word"))
    tg.set_outputs(tg_out=box["out"])

    assert check_inlined(tg.inline_boxes())

    nested = TierkreisGraph()
    box2 = nested.add_box(tg)
    nested.set_outputs(nested_out=box2["tg_out"])

    assert not check_inlined(nested.inline_boxes())

    assert check_inlined(nested.inline_boxes(True))


# test of bugfix
def test_insert_graph_id():
    tg_box = TierkreisGraph()
    tg_box.set_outputs(a=tg_box.input["a"], b=tg_box.input["b"])

    tg = TierkreisGraph()
    con = tg.add_const("word")
    outs = tg.insert_graph(tg_box, a=con, b=tg.input["b"])
    tg.set_outputs(**outs)

    # before bugfix outs was empty

    assert outs == {"a": con, "b": tg.input["b"]}
